﻿//The ALNS/TPF algorithm
//Authors: Lei He (l.he@tudelft.nl), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
//Date: June 3, 2019
//License: CC-BY-NC

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;

namespace GeneralTSALNS
{
    class ALNSTPF
    {
        /// <summary>
        /// The number of iterations for tabu heuristics
        /// </summary>
        int tabutime;
        /// <summary>
        /// The problem
        /// </summary>
        public problem Problem = new problem();
        /// <summary>
        /// The list of removal operators
        /// </summary>
        public List<Heuristic> DestroyList = new List<Heuristic>();
        /// <summary>
        /// The list of insertion operators
        /// </summary>
        public List<Heuristic> InsertList = new List<Heuristic>();
        /// <summary>
        /// A random value used in the randomized neighborhood operators
        /// </summary>
        public Random r = new Random(System.DateTime.Now.Millisecond);

        /// <summary>
        /// Remove jobs from the current solution and create a new solution (used by the removal operators)
        /// </summary>
        /// <param name="cursol">current solution</param>
        /// <param name="newsol">new solution</param>
        public void CheckRemove(List<SelectedOrder> cursol, List<SelectedOrder> newsol)
        {
            newsol.Clear();
            for (int i = 0; i < cursol.Count; i++)
            {
                if (!cursol[i].ToBeRemoved)
                    newsol.Add(cursol[i]);
                else
                {
                    cursol[i].ToBeRemoved = false;
                    if (Problem.IsInsertTabu)
                        cursol[i].Order.InsertTabu = r.Next(tabutime);
                }
            }
        }

        /// <summary>
        /// Return all the possible insertion positions of the candidate order
        /// </summary>
        /// <param name="OldSolution">Current solution</param>
        /// <param name="ObsInsert">Candidate order</param>
        /// <param name="Time">The start time</param>
        /// <returns>All possible positions</returns>
        public List<position> GetPositionListOfInsert(List<SelectedOrder> OldSolution, SelectedOrder ObsInsert, double Time)
        {
            List<position> positionlist = new List<position>(1);
            double TimePoint = Time;
            double TimePointEnd = 0;
            double folTimePoint = 0;
            double smalltrans = double.MaxValue;

            for (int i = 1; i < OldSolution.Count; i++)
            {
                if (OldSolution[i].Order.Deadline > Time)
                {
                    Time = OldSolution[i - 1].RealEndTime;
                    if (Time > ObsInsert.Order.Deadline)
                        break;

                    TimePoint = Problem.GetEarliestTime(OldSolution[i - 1], ObsInsert);
                    TimePointEnd = TimePoint + ObsInsert.Order.ProcessTime;
                    if (TimePointEnd > OldSolution[i].Order.Deadline||TimePoint > ObsInsert.Order.Deadline)
                        continue;
                    folTimePoint = Problem.GetTempEarliestTime(TimePointEnd, ObsInsert, OldSolution[i]);

                    if (folTimePoint - OldSolution[i].RealStartTime > OldSolution[i].TS)//Time slack is not enough
                        continue;

                    position tempposi = new position();
                    tempposi.posi = i;

                    tempposi.trans = (Problem.GetSetupTime(ObsInsert, OldSolution[i], TimePointEnd, folTimePoint)
                     + Problem.GetSetupTime(OldSolution[i - 1], ObsInsert, OldSolution[i - 1].RealEndTime, TimePoint)
                     - Problem.GetSetupTime(OldSolution[i - 1], OldSolution[i], OldSolution[i - 1].RealEndTime, OldSolution[i].RealStartTime));

                    tempposi.heuristic = tempposi.trans;

                    if (Problem.IsDueTime)
                        if ((folTimePoint - OldSolution[i].RealStartTime > OldSolution[i].DueTS && i != OldSolution.Count - 1) || TimePoint + ObsInsert.Order.ProcessTime > ObsInsert.Order.DueTime)
                        {
                            tempposi.heuristic = double.MaxValue;
                        }

                    if (tempposi.heuristic < double.MaxValue && tempposi.heuristic < smalltrans)
                    {
                        smalltrans = tempposi.heuristic;
                        positionlist.Clear();
                        positionlist.Add(tempposi);
                    }
                    if (tempposi.heuristic == double.MaxValue && smalltrans == double.MaxValue)
                        positionlist.Add(tempposi);
                }
            }
                
            return positionlist;
        }

        /// <summary>
        /// Insert a job back to the solution (used by the insertion operators: the fast insertion algorithm)
        /// </summary>
        /// <param name="OldSolution">current solution</param>
        /// <param name="ObsInsert">candidate job for the insertion</param>
        /// <param name="CurRev">the revenue of the current solution</param>
        /// <param name="NewRev">the revenue after the insertion</param>
        /// <returns></returns>
        public double InsertObsIntoSolWithTabu(ref List<SelectedOrder> OldSolution, SelectedOrder ObsInsert, double CurRev, out double NewRev)
        {
            List<position> positionlist = GetPositionListOfInsert(OldSolution, ObsInsert, ObsInsert.Order.Release);
            
            NewRev = CurRev;

            if (positionlist.Count == 0)
                return 0;

            double DeltaTrans = 0;//The increased setup time
            int k = -100;
            double OldStart = 0;
            double OldSetup = 0;

            if (positionlist[0].heuristic < double.MaxValue) //The insertion will not reduce the profit of other orders
            {
                k = positionlist[0].posi;
                OldSetup = OldSolution[k].Setup;
                OldSolution.Insert(k, ObsInsert);

                for (int j = k; j < OldSolution.Count; j++)//Update the start times and time slacks
                {
                    OldStart = OldSolution[j].RealStartTime;
                    OldSolution[j].RealStartTime = Problem.GetEarliestTime(OldSolution[j - 1], OldSolution[j]);
                    OldSolution[j].Setup = Problem.GetSetupTime(OldSolution[j - 1], OldSolution[j], OldSolution[j - 1].RealEndTime, OldSolution[j].RealStartTime);
                    if (OldSolution[j].RealStartTime == OldStart && j != k)
                        break;
                    OldSolution[j].TS -= (OldSolution[j].RealStartTime - OldStart);
                    OldSolution[j].DueTS -= (OldSolution[j].RealStartTime - OldStart);
                    OldSolution[j].RealEndTime = OldSolution[j].RealStartTime + OldSolution[j].Order.ProcessTime;
                }

                int flag = k;
                double SucStart = OldSolution[flag + 1].RealStartTime + OldSolution[flag + 1].TS;
                double SucDueStart = OldSolution[flag + 1].RealStartTime + OldSolution[flag + 1].DueTS;
                while (flag > 0)
                {
                    SucStart = Problem.GetLatestTime(SucStart, OldSolution[flag], OldSolution[flag + 1]);
                    double oldts = OldSolution[flag].TS;
                    double olddts = OldSolution[flag].DueTS;
                    OldSolution[flag].TS = Math.Min(OldSolution[flag].Order.Deadline - OldSolution[flag].RealStartTime,
                        SucStart - OldSolution[flag].RealStartTime);
                    if (Problem.IsDueTime)
                    {
                        SucDueStart = Problem.GetDueLatestTime(SucDueStart, OldSolution[flag], OldSolution[flag + 1]);
                        OldSolution[flag].DueTS = Math.Min(OldSolution[flag].Order.DueTime - OldSolution[flag].RealEndTime,
                            SucDueStart - OldSolution[flag].RealStartTime);
                    }
                    if (oldts == OldSolution[flag].TS && flag != k && olddts == OldSolution[flag].DueTS)
                    {
                        break;
                    }
                    flag--;
                }

                DeltaTrans = OldSolution[k].Setup + OldSolution[k + 1].Setup - OldSetup;
                NewRev += ObsInsert.Order.Revenue;
            }
            else //The insertion will reduce the profit of other orders because of the due time
            {
                double BestRev = CurRev;
                double BestDeltaTrans = 0;
                List<SelectedOrder> BestTempSol = new List<SelectedOrder>();
                for (int i = 0; i < positionlist.Count;i++ )
                {
                    k = positionlist[i].posi;
                    OldSetup = OldSolution[k].Setup;

                    List<SelectedOrder> TempSol = new List<SelectedOrder>();
                    CloneSol(OldSolution, TempSol);
                    TempSol.Insert(k, ObsInsert);
                    for (int j = k; j < TempSol.Count; j++)
                    {
                        OldStart = TempSol[j].RealStartTime;
                        TempSol[j].RealStartTime = Problem.GetEarliestTime(TempSol[j - 1], TempSol[j]);
                        TempSol[j].Setup = Problem.GetSetupTime(TempSol[j - 1], TempSol[j], TempSol[j - 1].RealEndTime, TempSol[j].RealStartTime);
                        if (TempSol[j].RealStartTime == OldStart && j != k)
                            break;
                        TempSol[j].RealEndTime = TempSol[j].RealStartTime + TempSol[j].Order.ProcessTime;
                    }
                    double TempFit = Problem.GetTotalProfit(TempSol);
                    if (TempFit > BestRev)
                    {
                        CloneSol(TempSol, BestTempSol);
                        BestRev = TempFit;
                        BestDeltaTrans = TempSol[k].Setup + TempSol[k + 1].Setup - OldSetup;
                    }
                }
                if (BestRev > CurRev)
                {
                    OldSolution = BestTempSol;
                    NewRev = BestRev;
                    AddTimeSlack(OldSolution);
                    DeltaTrans = BestDeltaTrans;
                }
            }
            return DeltaTrans;    
        }

        /// <summary>
        /// clone a sol which has the same property as the previous sol (but updating the new solution does not change the previous solution)
        /// </summary>
        /// <param name="A">current solution</param>
        /// <param name="B">the cloned new solution</param>
        public void CloneSol(List<SelectedOrder> A, List<SelectedOrder> B)
        {
            B.Clear();
            foreach (SelectedOrder a in A)
            {
                SelectedOrder temp = new SelectedOrder();
                temp.DueTS = a.DueTS;
                temp.Order = a.Order;
                temp.Profit = a.Profit;
                temp.RealEndTime = a.RealEndTime;
                temp.RealStartTime = a.RealStartTime;
                temp.Setup = a.Setup;
                temp.TS = a.TS;
                B.Add(temp);
            }
        }

        /// <summary>
        /// Main search process
        /// </summary>
        /// <param name="time">the accumulated running time for each iteration</param>
        /// <param name="rev">the best revenue for each iteration</param>
        public void StartExecute(out List<double> time, out List<double> rev)
        {
            #region PreProcessing
            int MaxIteration = Problem.MaxIteration;
            double shortprocess = double.MaxValue;
            double TotalRevenue = 0, TotalUnitRevenue = 0, TotalSetup = 0, TotalConflict = 0;
            for (int i = 0; i < Problem.OrderList.Count; i++)
            {
                TotalRevenue += Problem.OrderList[i].Revenue;
                TotalUnitRevenue += Problem.OrderList[i].UnitRevenue;
                TotalSetup += Problem.OrderList[i].TotalSetup;
                TotalConflict += Problem.OrderList[i].Conflict;
                if (Problem.OrderList[i].ProcessTime < shortprocess && Problem.OrderList[i].ProcessTime != 0)
                    shortprocess = Problem.OrderList[i].ProcessTime;
            }
            
            int InnerIteration = 50;
            int SigmaGlobal = 30;
            int SigmaLocal = 20;
            int SigmaBad = 10;
            double Temperature = 100;
            double ParaOfAnnealing = 0.99975;
            double per = 0.1;
            int SizeOfRequestBank = 1;
            double Lam = 0.5;
            bool IsEnd = false;

            Heuristic CurDelete = new Heuristic();
            CurDelete.type = "Destroy";
            Heuristic CurInsert = new Heuristic();
            CurInsert.type = "Repair";
            int SelectDelNum, SelectInsNum;
            
            time = new List<double>(MaxIteration);
            rev = new List<double>(MaxIteration);
            tabutime = (int)Math.Sqrt(Problem.OrderList.Count / 2);
            InitialOperatorPool();

            List<SelectedOrder> CurrentSol = new List<SelectedOrder>(Problem.OrderList.Count);
            List<SelectedOrder> GlobalOptimum = new List<SelectedOrder>(Problem.OrderList.Count);
            List<SelectedOrder> NewCurrent = new List<SelectedOrder>(Problem.OrderList.Count);
            List<SelectedOrder> BankList = new List<SelectedOrder>(Problem.OrderList.Count);
            #endregion

            Stopwatch st = new Stopwatch();
            st.Start();

            CurrentSol = InitialSolution();
            CopyListToList(CurrentSol, GlobalOptimum);

            double C_fit = Problem.GetTotalProfit(@CurrentSol);
            double G_fit = C_fit;
            double NC_fit = 0;

            Temperature = -(0.05 / Math.Log(0.5) * C_fit);
            if (Temperature > 100)
                Temperature = 100;
            double initemperature = Temperature;
           
            for (int i = 0; i < MaxIteration / InnerIteration; i++)
            {
                for (int j = 0; j < InnerIteration; j++)
                {
                    SizeOfRequestBank = (int)Math.Ceiling(CurrentSol.Count * per);

                    ////***********Tabu search************//
                    for (int ii = 0; ii < Problem.OrderList.Count; ii++)
                    {
                        Problem.OrderList[ii].RemoveTabu--;
                        Problem.OrderList[ii].InsertTabu--;
                    }
                    ////***********Tabu search************//    

                    rev.Add(G_fit);
                    time.Add(st.Elapsed.TotalSeconds);

                    SelectDelNum = SelectOfHeuristic(DestroyList);
                    CurDelete = DestroyList[SelectDelNum];
                    CurDelete.CalledTimes++;

                    CurDelete.DestroyHeu(CurDelete.Heur_Id, CurrentSol, SizeOfRequestBank, r);

                    for (int ii = 0; ii < Problem.SelectedOrderList.Count; ii++)
                    {
                        Problem.SelectedOrderList[ii].Order.IsInNewCurrent = false;
                        Problem.SelectedOrderList[ii].Order.FatherOrder.IsInCompound = false;
                        Problem.SelectedOrderList[ii].Order.FatherOrder.IsProcessed = false;
                    }

                    CheckRemove(CurrentSol, NewCurrent);

                    for (int ii = 0; ii < NewCurrent.Count; ii++)
                    {
                        SelectedOrder tempCurrent = new SelectedOrder();
                        CloneOrder(NewCurrent[ii], tempCurrent);
                        NewCurrent[ii] = tempCurrent;
                        NewCurrent[ii].Order.IsInNewCurrent = true;
                        NewCurrent[ii].Order.FatherOrder.IsProcessed = true;
                    }

                    if (NewCurrent.Count > 0)
                        UpdateSolution(ref NewCurrent);

                    double CurrentRepairRevenue = 0;
                    for (int ii = 0; ii < NewCurrent.Count; ii++)
                        CurrentRepairRevenue += NewCurrent[ii].Order.Revenue;
                    double BankRevenue = 0;

                    BankList.Clear();

                    for (int ii = 0; ii < Problem.SelectedOrderList.Count; ii++)
                        if (!Problem.SelectedOrderList[ii].Order.IsInNewCurrent && !Problem.SelectedOrderList[ii].Order.FatherOrder.IsProcessed && Problem.SelectedOrderList[ii].Order.InsertTabu < 0)
                        {
                            BankList.Add(Problem.SelectedOrderList[ii]);
                            BankRevenue += Problem.SelectedOrderList[ii].Order.Revenue;
                        }

                    double IdleTime = Problem.OrderList[Problem.OrderList.Count - 1].Deadline - Problem.OrderList[0].Release;
                    for (int idle = 0; idle < NewCurrent.Count - 1; idle++)
                    {
                        IdleTime -= (NewCurrent[idle + 1].Setup + NewCurrent[idle + 1].Order.ProcessTime);
                    }

                    //**************Repair solution***********************
                    SelectInsNum = SelectOfHeuristic(InsertList);
                    CurInsert = InsertList[SelectInsNum];
                    CurInsert.CalledTimes++;

                    switch (CurInsert.Heur_Id)
                    {
                        case 0:
                            for (int inser = 0; inser < BankList.Count; inser++)
                                BankList[inser].HeuristicInfo = BankList[inser].Order.Revenue * (1 + r.NextDouble());
                            BankList = BankList.OrderByDescending(s => s.HeuristicInfo).ToList<SelectedOrder>();
                            break;
                        case 1:
                            for (int inser = 0; inser < BankList.Count; inser++)
                                BankList[inser].HeuristicInfo = BankList[inser].Order.UnitRevenue * (1 + r.NextDouble());
                            BankList = BankList.OrderByDescending(s => s.HeuristicInfo).ToList<SelectedOrder>();
                            break;
                        case 2:
                            for (int inser = 0; inser < BankList.Count; inser++)
                                BankList[inser].HeuristicInfo = BankList[inser].Order.TotalSetup * (1 + r.NextDouble());
                            BankList = BankList.OrderBy(s => s.HeuristicInfo).ToList<SelectedOrder>();
                            break;
                        case 3:
                            for (int inser = 0; inser < BankList.Count; inser++)
                                BankList[inser].HeuristicInfo = BankList[inser].Order.FatherOrder.VTWList.Count * (1 + r.NextDouble());
                            BankList = BankList.OrderBy(s => s.HeuristicInfo).ToList<SelectedOrder>();
                            break;
                        case 4:
                            for (int inser = 0; inser < BankList.Count; inser++)
                                BankList[inser].HeuristicInfo = BankList[inser].Order.Conflict * (1 + r.NextDouble());
                            BankList = BankList.OrderBy(s => s.HeuristicInfo).ToList<SelectedOrder>();
                            break;
                    }

                    if (Problem.IsInsertTabu)
                    {
                        List<SelectedOrder> TempBankList = new List<SelectedOrder>();
                        for (int ins = 0; ins < BankList.Count; ins++)
                            if (BankList[ins].Order.InsertTabu >= 0)
                            {
                                TempBankList.Add(BankList[ins]);
                                BankList.Remove(BankList[ins]);
                                ins--;
                            }
                        for (int ins = 0; ins < TempBankList.Count; ins++)
                            BankList.Add(TempBankList[ins]);
                    }

                    bool jump = false;
                    SelectedOrder tempObs = new SelectedOrder();

                    for (int k = 0; k < BankList.Count; k++)
                    {
                        tempObs = BankList[k];
                        BankList.Remove(tempObs);
                        BankRevenue -= tempObs.Order.Revenue;

                        if (CurrentRepairRevenue + tempObs.Order.Revenue + BankRevenue < C_fit)
                        {
                            jump = true;
                            break;
                        }

                        if (IdleTime < tempObs.Order.ProcessTime || tempObs.Order.FatherOrder.IsProcessed)
                        {
                            if (IdleTime < shortprocess)
                                break;
                            k = -1;
                            continue;
                        }

                        int IniLength = NewCurrent.Count;
                        SelectedOrder SelectedObs = new SelectedOrder();
                        CloneOrder(tempObs, SelectedObs);
                        double NewRev = 0;
                        double DeltaTrans = InsertObsIntoSolWithTabu(ref NewCurrent, SelectedObs, CurrentRepairRevenue,out NewRev); 
                        if (NewCurrent.Count > IniLength)
                        {
                            SelectedObs.Order.IsInNewCurrent = true;
                            SelectedObs.Order.FatherOrder.IsProcessed = true;
                            IdleTime -= (SelectedObs.Order.ProcessTime + DeltaTrans);
                            CurrentRepairRevenue = NewRev;
                            if (Problem.IsRemoveTabu)
                                SelectedObs.Order.RemoveTabu = r.Next(tabutime);
                        }
                        k = -1;
                    }
                    if (jump)
                        continue;

                    NC_fit = Problem.GetTotalProfit(NewCurrent);

                    double Com_fit;
                    NewCurrent = GenerateCompoundSolution(CurrentSol, NewCurrent, NC_fit, out Com_fit, 10);
                    NC_fit = Com_fit;

                    #region Update Solution
                    if ((NC_fit > G_fit) || (NC_fit == G_fit && NewCurrent[NewCurrent.Count - 2].RealEndTime < GlobalOptimum[GlobalOptimum.Count - 2].RealEndTime))
                    {
                        CopyListToList(NewCurrent, CurrentSol);
                        C_fit = NC_fit;
                        CopyListToList(NewCurrent, GlobalOptimum);
                        G_fit = NC_fit;
                        CurDelete.score += SigmaGlobal;
                        CurInsert.score += SigmaGlobal;
                    }
                    else
                    {
                        if (NC_fit >= C_fit)
                        {
                            CopyListToList(NewCurrent, CurrentSol);
                            C_fit = NC_fit;
                            CurDelete.score += SigmaLocal;
                            CurInsert.score += SigmaLocal;
                        }
                        else
                        {
                            double pp = Math.Exp(100 / Temperature * ((NC_fit - C_fit) / C_fit));
                            double temp = r.NextDouble();
                            if (pp >= temp)
                            {
                                CopyListToList(NewCurrent, CurrentSol);
                                C_fit = NC_fit;
                                CurDelete.score += SigmaBad;
                                CurInsert.score += SigmaBad;
                            }
                        }
                    }

                    Temperature = Temperature * ParaOfAnnealing;
                    #endregion        
                    
                    if (G_fit== TotalRevenue)
                    {
                        IsEnd = true;
                        break;
                    }        
                }
                if (IsEnd)
                    break;

                UpdateHeuristicList(DestroyList, Lam);
                UpdateHeuristicList(InsertList, Lam);    
            }

            st.Stop();

            rev.Add(G_fit);
            time.Add(st.Elapsed.TotalSeconds);

            bool Feasible = Problem.CheckSolution(GlobalOptimum);

            if (!Feasible)
                rev[rev.Count - 1] = -999999;
        }

        /// <summary>
        /// Update the weights of the operators
        /// </summary>
        /// <param name="NameOfList"></param>
        /// <param name="Lam"></param>
        public void UpdateHeuristicList(List<Heuristic> NameOfList, double Lam)
        {
            double total = 0;
            int M = NameOfList.Count;

            for (int i = 0; i < NameOfList.Count; i++)
            {
                if (NameOfList[i].CalledTimes != 0)
                    NameOfList[i].score = 1 + (NameOfList[i].score - 1) / NameOfList[i].CalledTimes;
                else
                    NameOfList[i].score = 1;
                total += NameOfList[i].score;
                NameOfList[i].CalledTimes = 0;
            }
            for (int i = 0; i < NameOfList.Count; i++)
            {
                double temp = NameOfList[i].score / total;
                NameOfList[i].weight = NameOfList[i].weight * Lam + (1 - Lam) * temp;
                NameOfList[i].score = 1;
            }
        }

        /// <summary>
        /// Select a neighborhood operator
        /// </summary>
        /// <param name="HeurList"></param>
        /// <returns></returns>
        public int SelectOfHeuristic(List<Heuristic> HeurList)
        {
            int HeuId = 0;
            double CurValue = r.NextDouble();
            double SumWeight = 0;
            for (int i = 0; i < HeurList.Count; i++)
            {
                SumWeight = SumWeight + HeurList[i].weight;
                if (CurValue <= SumWeight)
                {
                    HeuId = i;
                    break;
                }
            }
            return HeuId;
        }

        /// <summary>
        /// Update the solution to start each order as early as possible and update the corresponding time slacks
        /// </summary>
        /// <param name="NewObsList"></param>
        public virtual void UpdateSolution(ref List<SelectedOrder> NewObsList)
        {
            for (int i = 1; i < NewObsList.Count - 1; i++)
            {
                NewObsList[i].RealStartTime = Problem.GetEarliestTime(NewObsList[i - 1], NewObsList[i]);
                if (NewObsList[i].RealStartTime > NewObsList[i].Order.Deadline)
                {
                    NewObsList.Remove(NewObsList[i]);
                    NewObsList[i].Order.IsInNewCurrent = false;
                    i--;
                    continue;
                }
                NewObsList[i].RealEndTime = NewObsList[i].RealStartTime + NewObsList[i].Order.ProcessTime;
                if (NewObsList[i].RealEndTime >= NewObsList[i + 1].Order.Deadline)
                {
                    NewObsList.Remove(NewObsList[i]);
                    NewObsList[i].Order.IsInNewCurrent = false;
                    i--;
                    continue;
                }
                NewObsList[i].Setup = Problem.GetSetupTime(NewObsList[i - 1], NewObsList[i], NewObsList[i - 1].RealEndTime, NewObsList[i].RealStartTime);
                if (NewObsList[i].RealEndTime + Problem.GetSetupTime(NewObsList[i], NewObsList[NewObsList.Count - 1], NewObsList[i].RealEndTime, NewObsList[NewObsList.Count - 1].RealStartTime) > NewObsList[NewObsList.Count - 1].Order.Deadline)
                {
                    NewObsList.Remove(NewObsList[i]);
                    NewObsList[i].Order.IsInNewCurrent = false;
                    i--;
                }
            }
            AddTimeSlack(NewObsList);
        }

        /// <summary>
        /// Deep copy an order
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        public virtual void CloneOrder(SelectedOrder a, SelectedOrder b)
        {
            b.TS = a.TS;
            b.DueTS = a.DueTS;
            b.Order = a.Order;
            b.Setup = a.Setup;
            b.Profit = a.Profit;
            b.JustInsert = a.JustInsert;
            b.RealEndTime = a.RealEndTime;
            b.RealStartTime = a.RealStartTime;
        }


        public void InitialOperatorPool()
        {
            for (int i = 0; i < Problem.DestroyList.Length; i++)
            {
                Heuristic a = new Heuristic();
                a.Heur_Id = Problem.DestroyList[i];
                decimal temp = (decimal)1 / Problem.DestroyList.Length; ;
                a.weight = (double)temp;
                a.type = "Destroy";
                a.score = 1;
                a.CalledTimes = 0;
                DestroyList.Add(a);
            }
            for (int i = 0; i < Problem.RepairList.Length; i++)
            {
                Heuristic a = new Heuristic();
                a.Heur_Id = Problem.RepairList[i];
                decimal temp = (decimal)1 / Problem.RepairList.Length; ;
                a.weight = (double)temp;
                a.type = "Repair";
                a.score = 1;
                a.CalledTimes = 0;
                InsertList.Add(a);
            }
        }

        /// <summary>
        /// Generalize the initial solution
        /// </summary>
        /// <returns></returns>
        public List<SelectedOrder> InitialSolution()
        {
            List<SelectedOrder> IniSolList = new List<SelectedOrder>(Problem.OrderList.Count);

            Problem.OrderList = Problem.OrderList.OrderBy(s => s.Release).ToList<Order>();

            Problem.SelectedOrderList = CreateSelectedOrderList(Problem.OrderList);

            IniSolList.Add(Problem.SelectedOrderList[0]);

            SelectedOrder curorder, preorder;

            for (int i = 1; i < Problem.OrderList.Count - 1; i++)
            {
                if (Problem.IsMultiVTW)
                    if (Problem.SelectedOrderList[i].Order.FatherOrder.IsProcessed)
                        continue;
                curorder = Problem.SelectedOrderList[i];
                preorder = IniSolList[IniSolList.Count - 1];

                double NeededTime = Problem.GetEarliestTime(preorder, curorder);

                if (NeededTime <= curorder.Order.Deadline)
                {
                    curorder.RealStartTime = NeededTime;
                    curorder.RealEndTime = NeededTime + curorder.Order.ProcessTime;
                    curorder.Setup = Problem.GetSetupTime(preorder, curorder, preorder.RealEndTime, curorder.RealStartTime);
                    curorder.Order.IsAccept = true;
                    IniSolList.Add(curorder);
                    if (Problem.IsMultiVTW)
                        curorder.Order.FatherOrder.IsProcessed = true;
                }
                else
                    curorder.Order.IsAccept = false;

            }

            while (true)
            {
                curorder = Problem.SelectedOrderList[Problem.SelectedOrderList.Count - 1];
                preorder = IniSolList[IniSolList.Count - 1];

                double Tran_time1 = Problem.GetSetupTime(preorder, curorder, preorder.RealEndTime, curorder.RealStartTime);
                if (preorder.RealEndTime + Tran_time1 > curorder.Order.Deadline)
                {
                    preorder.Order.IsAccept = false;
                    IniSolList.Remove(preorder);
                }
                else
                    break;
            }

            IniSolList.Add(Problem.SelectedOrderList[Problem.SelectedOrderList.Count - 1]);
            IniSolList = AddTimeSlack(IniSolList);
            return IniSolList;
        }

        public void CopyListToList(List<SelectedOrder> a, List<SelectedOrder> b)
        {
            if (b.Count != 0)
                b.Clear();
            for (int i = 0; i < a.Count; i++)
                b.Add(a[i]);
        }

        /// <summary>
        /// Generate the compound solution, and if it is better than the new solution but not the same as the current solution, it is used as the new solution
        /// </summary>
        /// <param name="CurSol"></param>
        /// <param name="NewSol"></param>
        /// <param name="Nfit"></param>
        /// <param name="ComRev"></param>
        /// <param name="Length">the number of jobs in a partial sequence</param>
        /// <returns></returns>
        public virtual List<SelectedOrder> GenerateCompoundSolution(List<SelectedOrder> CurSol, List<SelectedOrder> NewSol, double Nfit, out double ComRev, int Length)
        {
            List<SelectedOrder> ComSol = new List<SelectedOrder>();
            List<SmallPart> SmallPartList = new List<SmallPart>();
            SmallPart sp = new SmallPart();
            SmallPartList.Add(sp);

            for (int i = 0; i < CurSol.Count; i++)
            {
                SmallPartList[SmallPartList.Count - 1].list.Add(CurSol[i]);
                if (SmallPartList[SmallPartList.Count - 1].list.Count == Length)
                {
                    SmallPartList[SmallPartList.Count - 1].EndTime = CurSol[i].RealEndTime;
                    SmallPart sp1 = new SmallPart();
                    SmallPartList.Add(sp1);
                }
            }
            SmallPartList[SmallPartList.Count - 1].EndTime = CurSol[CurSol.Count - 1].RealEndTime;

            int count = 0, NewSolCount = 0, CurSolCount = 0;
            SmallPart newsp = new SmallPart();
            SmallPart cursp = new SmallPart();
            bool jump = false;

            while (count < SmallPartList.Count)
            {
                while (NewSol[NewSolCount].RealEndTime <= SmallPartList[count].EndTime)
                {
                    if (!NewSol[NewSolCount].Order.FatherOrder.IsInCompound)
                    {
                        newsp.list.Add(NewSol[NewSolCount]);
                        newsp.Revenue += NewSol[NewSolCount].Order.Revenue;
                    }
                    NewSolCount++;
                    if (NewSolCount == NewSol.Count - 1)
                    {
                        jump = true;
                        break;
                    }
                }

                while (CurSol[CurSolCount].RealEndTime <= SmallPartList[count].EndTime)
                {
                    if (!CurSol[CurSolCount].Order.FatherOrder.IsInCompound)
                    {
                        cursp.list.Add(CurSol[CurSolCount]);
                        cursp.Revenue += CurSol[CurSolCount].Order.Revenue;
                    }
                    CurSolCount++;
                    if (CurSolCount == CurSol.Count - 1)
                    {
                        jump = true;
                        break;
                    }
                }

                if (newsp.Revenue > cursp.Revenue)
                {
                    for (int j = 0; j < newsp.list.Count; j++)
                    {
                        newsp.list[j].Order.FatherOrder.IsInCompound = true;
                        SelectedOrder temp = new SelectedOrder();
                        CloneOrder(newsp.list[j], temp);
                        ComSol.Add(temp);
                    }
                }
                else
                {
                    for (int j = 0; j < cursp.list.Count; j++)
                    {
                        cursp.list[j].Order.FatherOrder.IsInCompound = true;
                        SelectedOrder temp = new SelectedOrder();
                        CloneOrder(cursp.list[j], temp);
                        ComSol.Add(temp);
                    }
                }
                count++;
                if (jump)
                    break;

                newsp.list.Clear();
                newsp.Revenue = 0;
                cursp.list.Clear();
                cursp.Revenue = 0;

            }

            while (NewSolCount < NewSol.Count - 1)
            {
                if (!NewSol[NewSolCount].Order.FatherOrder.IsInCompound)
                {
                    NewSol[NewSolCount].Order.FatherOrder.IsInCompound = true;
                    SelectedOrder temp = new SelectedOrder();
                    CloneOrder(NewSol[NewSolCount], temp);
                    ComSol.Add(temp);
                }
                NewSolCount++;
            }
            while (CurSolCount < CurSol.Count - 1)
            {
                if (!CurSol[CurSolCount].Order.FatherOrder.IsInCompound)
                {
                    CurSol[CurSolCount].Order.FatherOrder.IsInCompound = true;
                    SelectedOrder temp = new SelectedOrder();
                    CloneOrder(CurSol[CurSolCount], temp);
                    ComSol.Add(temp);
                }
                CurSolCount++;
            }
            ComSol.Add(CurSol[CurSol.Count - 1]);
            UpdateSolution(ref ComSol);
            ComRev = Problem.GetTotalProfit(ComSol);

            if (ComRev > Nfit)
            {
                return ComSol;
            }
            else
            {
                ComRev = Nfit;
                return NewSol;
            }
        }

        /// <summary>
        /// Generate a list of scheduled orders. Note that not all of these orders will be included in the solution
        /// </summary>
        /// <param name="ol"></param>
        /// <returns></returns>
        public List<SelectedOrder> CreateSelectedOrderList(List<Order> ol)
        {
            List<SelectedOrder> List = new List<SelectedOrder>(Problem.OrderList.Count);

            for (int i = 0; i < ol.Count; i++)
            {
                SelectedOrder Cur_obs = new SelectedOrder();
                Cur_obs.RealStartTime = ol[i].Release; 
                Cur_obs.RealEndTime = Cur_obs.RealStartTime + ol[i].ProcessTime;
                Cur_obs.Order = ol[i];
                List.Add(Cur_obs);
            }
            return List;
        }

        /// <summary>
        /// Calculate the time slack and due time slack of an order
        /// </summary>
        /// <param name="ListWithoutTS"></param>
        /// <returns></returns>
        public virtual List<SelectedOrder> AddTimeSlack(List<SelectedOrder> ListWithoutTS)
        {
            int i = ListWithoutTS.Count - 1;
            ListWithoutTS[i].TS = ListWithoutTS[i].Order.Deadline - ListWithoutTS[i].RealStartTime;
            if (Problem.IsDueTime)
                ListWithoutTS[i].DueTS = ListWithoutTS[i].Order.DueTime - ListWithoutTS[i].RealEndTime;
            double SucStart = ListWithoutTS[i].Order.Deadline;
            double SucDueStart = ListWithoutTS[i].Order.DueTime - ListWithoutTS[i].Order.ProcessTime;
            i--;
            while (i >= 0)
            {
                SucStart = Problem.GetLatestTime(SucStart, ListWithoutTS[i], ListWithoutTS[i + 1]);
                ListWithoutTS[i].TS = Math.Min(ListWithoutTS[i].Order.Deadline - ListWithoutTS[i].RealStartTime,
                    SucStart - ListWithoutTS[i].RealStartTime);

                if (Problem.IsDueTime)
                {
                    SucDueStart = Problem.GetDueLatestTime(SucDueStart, ListWithoutTS[i], ListWithoutTS[i + 1]);
                    ListWithoutTS[i].DueTS = Math.Min(ListWithoutTS[i].Order.DueTime - ListWithoutTS[i].RealEndTime,
                        SucDueStart - ListWithoutTS[i].RealStartTime);
                }
                i--;
            }
            return ListWithoutTS;
        }

    }
}



